/* -*- mode: c++; tab-width: 8; c-file-style: "nester" -*- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "debug.h"
#include "nes.h"
#include "null_sound_mgr.h"
#include "sdl_sound_mgr.h"
#include "sdl_saving.h"
#include "sdl_gg.h"

extern "C" {
#include "rc.h"
#include "input.h"
}

#define CMDLINE_FILE "/tmp/znester-cmdline"

/* exports.c */
extern "C" {
void init_exports(void);
void show_exports(void);

int rc_command(char *line);
void rc_dokey(int key, int st);
int splitline(char **argv, int max, char *line);
}

static int
mergeCmdlineFile(const char *filename, int orig_argc, char **orig_argv,
                 int *argc, char **argv);
static int
parseArgs(int argc, char **argv);

void doevents(void);

NES *g_emu;

//
static char *st_saveDir;
static char *st_snapDir;

static char *defaultconfig[] =
{
  "bind esc quit",

  "bind up +up",
  "bind q +up",
  "bind down +down",
  "bind shift +down",
  "bind left +left",
  "bind tab +left",
  "bind right +right",
  "bind a +right",
  "bind x +a",
  "bind z +b",
  "bind l +a",
  "bind k +b",
  "bind enter +start",
  "bind j +start",
  "bind space +select",
  "bind h +select",

  "bind penup +up",
  "bind penright +right",
  "bind pendown +down",
  "bind penleft +left",
  "bind pencenterright +a",
  "bind pencenterleft +b",

  "bind 1 \"diskside 1\"",
  "bind 2 \"diskside 2\"",
  "bind 3 \"diskside 3\"",
  "bind 4 \"diskside 4\"",
  "bind 7 savestate",
  "bind 8 decsaveslot",
  "bind 9 incsaveslot",
  "bind 0 loadstate",
  "bind r reset",

//   "set vmode 320 240 16",
//   "set stereo false",
  "set grabinput true",

  "source znester.rc",
  NULL
};

static void banner(void)
{
  printf("\nnester " "7.1"/*VERSION*/ "\n");
}

static void copyright(void)
{
  banner();
  printf(
    "Copyright (C) 2000-2001 AUTHOR\n"
    "Portions contributed by other authors; see CREDITS for details.\n"
    "\n"
    "This program is free software; you can redistribute it and/or modify\n"
    "it under the terms of the GNU General Public License as published by\n"
    "the Free Software Foundation; either version 2 of the License, or\n"
    "(at your option) any later version.\n"
    "\n"
    "This program is distributed in the hope that it will be useful,\n"
    "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
    "GNU General Public License for more details.\n"
    "\n"
    "You should have received a copy of the GNU General Public License\n"
    "along with this program; if not, write to the Free Software\n"
    "Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n"
    "\n");
}

static void usage(char *name)
{
  copyright();
  printf("Type %s --help for detailed help.\n\n", name);
  exit(1);
}

static void copying(void)
{
  copyright();
  exit(0);
}

static void help(char *name)
{
  banner();
  printf("Usage: %s [options] romfile\n", name);
  printf("\n"
         "      --source FILE             read rc commands from FILE\n"
         "      --bind KEY COMMAND        bind KEY to perform COMMAND\n"
         "      --VAR=VALUE               set rc variable VAR to VALUE\n"
         "      --VAR                     set VAR to 1 (turn on boolean options)\n"
         "      --no-VAR                  set VAR to 0 (turn off boolean options)\n"
         "      --showvars                list all available rc variables\n"
         "      --help                    display this help and exit\n"
         "      --version                 output version information and exit\n"
         "      --copying                 show copying permissions\n"
         "");
  exit(0);
}

static void version(char *name)
{
  printf("%s-" "7.1"/*VERSION*/ "\n", name);
  exit(0);
}

bool exit_game_loop = false;
bool interrupt_game_loop = false;

static uint32 frame_count = 0;
static volatile int rest_frames = 0;
static SDL_Surface *screen;

bool skip_sound = false;
static bool draw_pattern = true;

char nes_homedir[256];

uint16 
get_nesscreen_pixel_color(int x, int y)
{
  /* NOT SUPPORTED */
  return 0; 
}


uint8
keymap_sdl2dc[256];

static void
keymap_sdl2dc_init()
{
  int i;
  for (i = 0; i < 256; ++i)
    keymap_sdl2dc[i] = NES_device_kbd::NES_KEY_VoidSymbol;
  
  for (i = 0; i < 26; ++i)
    keymap_sdl2dc[SDLK_a + i] = NES_device_kbd::NES_KEY_A + i;
  
  for (i = 0; i < 10; ++i)
    keymap_sdl2dc[SDLK_0 + i] = NES_device_kbd::NES_KEY_0 + i;
}


void
draw_pattern_table(NES *emu)
{
  uint8 *p;
  uint16 *base;
  uint16 *buf;
  uint32 pattern1;
  uint32 pattern2;
  int xx, yy, x, y;
  int bank;
  int i;
  uint16* pattern_palette = emu->ppu_bg_pal_color ();
  
  if (SDL_MUSTLOCK (screen)) SDL_LockSurface (screen);
  
  base = (uint16 *)(screen->pixels) + 320 * 240;
  for (i = 0; i < 320; ++i)
  {
    *base++ = 0xffff;
  }
  
  for (bank = 0; bank < 8; ++bank)
  {
    p = emu->ppu_vram_banks(bank);
    
    for (yy = 0; yy < 2; ++yy)
    {
      for (xx = 0; xx < 32; ++xx)
      {
	for (y = 0; y < 8; ++y)
	{
	  pattern1 = (*(p + 8) & 0xaa) | ((*p >> 1) & 0x55);
	  pattern2 = ((*(p + 8) << 1) & 0xaa) | (*p & 0x55);
	  ++p;
	  
	  buf = base + 32 + (bank * 320 * 16) + (yy * 320 * 8) + (xx * 8) + (y * 320);
	  
	  buf += 8;
	  *(--buf) = pattern_palette[(pattern2 & 0x03)];
	  *(--buf) = pattern_palette[(pattern1 & 0x03)];
	  pattern2 >>= 2;
	  pattern1 >>= 2;
	  
	  *(--buf) = pattern_palette[(pattern2 & 0x03)];
	  *(--buf) = pattern_palette[(pattern1 & 0x03)];
	  pattern2 >>= 2;
	  pattern1 >>= 2;
	  
	  *(--buf) = pattern_palette[(pattern2 & 0x03)];
	  *(--buf) = pattern_palette[(pattern1 & 0x03)];
	  pattern2 >>= 2;
	  pattern1 >>= 2;
	  
	  *(--buf) = pattern_palette[(pattern2 & 0x03)];
	  *(--buf) = pattern_palette[(pattern1 & 0x03)];
	}
	p += 8;
      }
    }
  }
  
finish:
  
  if (SDL_MUSTLOCK (screen)) SDL_UnlockSurface (screen);
}


static void
wait_until_release_button (SDLKey key)
{
  uint8 *keys;
  
  SDL_PumpEvents ();
  keys = SDL_GetKeyState (0);
  while (keys[key])
  {
    SDL_Delay (50);
    SDL_PumpEvents ();
    keys = SDL_GetKeyState (0);
  }
}


static void
poll_input(NES *emu)
{
  static int pad_num = 0;
  static int snap_num = 0;
  uint8 *keys;
  NES_pad *pad = emu->get_pad (pad_num);

  SDL_PumpEvents ();
  keys = SDL_GetKeyState (0);
  
  if (keys[SDLK_ESCAPE])
  {
    exit_game_loop = true;
    interrupt_game_loop = true;
  }
  
  if (keys[SDLK_r] && keys[SDLK_LSHIFT])
  {
    emu->reset ();
    LOG("reset");
    wait_until_release_button (SDLK_r);
  }
  
  if (keys[SDLK_p])
  {
    LOG ("pause");
    wait_until_release_button (SDLK_p);
    
    while (!keys[SDLK_p])
    {
      SDL_Delay (10);
      SDL_PumpEvents ();
      keys = SDL_GetKeyState (0);
    }
    
    LOG ("quit pause");
    wait_until_release_button (SDLK_p);
  }
  
  if (keys[SDLK_1])
  {
    emu->SetDiskSide (0x01);
    LOG("set diskside 1A");
    wait_until_release_button (SDLK_1);
  }
  
  if (keys[SDLK_2])
  {
    emu->SetDiskSide (0x02);
    LOG("set diskside 1B");
    wait_until_release_button (SDLK_2);
  }
  
  if (keys[SDLK_3])
  {
    emu->SetDiskSide (0x03);
    LOG("set diskside 2A");
    wait_until_release_button (SDLK_3);
  }
  
  if (keys[SDLK_4])
  {
    emu->SetDiskSide (0x04);
    LOG("set diskside 2B");
    wait_until_release_button (SDLK_4);
  }
  
  if (keys[SDLK_5])
  {
    char filename[512];
    
    snprintf(filename, sizeof(filename),
	     "%s/snap/%s%d", 
             nes_homedir,
	     emu->getROMname (),
	     snap_num);
    if (emu->saveState(filename))
    {
      LOG("snapsave to");
      LOG(filename);
    }
    wait_until_release_button (SDLK_5);
  }
  
  if (keys[SDLK_7])
  {
    char filename[512];
    
    snprintf(filename, sizeof(filename),
	     "%s/snap/%s%d", 
             nes_homedir,
	     emu->getROMname (), 
	     snap_num);
    if (emu->loadState(filename))
    {
      LOG("snapload from");
      LOG(filename);
    }
    wait_until_release_button (SDLK_7);
  }
  
  if (keys[SDLK_w])
  {
    if (snap_num == 0)
      snap_num = 8;
    else
      --snap_num;
    
    printf("snap_num = %d\n", snap_num);
    wait_until_release_button (SDLK_w);
  }
  
  if (keys[SDLK_e])
  {
    if (snap_num == 8)
      snap_num = 0;
    else
      ++snap_num;
    
    printf("snap_num = %d\n", snap_num);
    wait_until_release_button (SDLK_e);
  }
  
#if 0
  if (keys[SDLK_EQUALS])
  {
    ++(emu->CYCLES_PER_LINE);
    
    printf("CYCLES_PER_LINE = %f\n", emu->CYCLES_PER_LINE);
    
    if (!keys[SDLK_LSHIFT])
      wait_until_release_button (SDLK_EQUALS);
  }
  if (keys[SDLK_MINUS])
  {
    --(emu->CYCLES_PER_LINE);
    
    printf("CYCLES_PER_LINE = %f\n", emu->CYCLES_PER_LINE);
    
    if (!keys[SDLK_LSHIFT])
      wait_until_release_button (SDLK_MINUS);
  }
#endif
  if (keys[SDLK_BACKSLASH])
  {
    pad_num = (pad_num + 1) & 0x03;
    
    printf("pad_num = %d\n", pad_num);
    
    wait_until_release_button (SDLK_BACKSLASH);
  }
  if (keys[SDLK_d])
  {
    draw_pattern = !draw_pattern;
    
    printf("draw_pattern = %d\n", draw_pattern);
    
    wait_until_release_button (SDLK_d);
  }
  
  {
    int key;

#if 1
    emu->NES_kbd.clear_register ();
#endif
    
    for (key = SDLK_FIRST; key < SDLK_LAST; ++key)
      if (keys[key]) emu->NES_kbd.setkey(keymap_sdl2dc[key]);

#if 0    
    emu->NES_kbd.setkey(NES_device_kbd::NES_KEY_M);
#endif
  }
  
  skip_sound = keys[SDLK_s];
//   skip_sound = 1;
  
  pad->nes_UP = keys[SDLK_UP];
  pad->nes_DOWN = keys[SDLK_DOWN];
  pad->nes_LEFT = keys[SDLK_LEFT];
  pad->nes_RIGHT = keys[SDLK_RIGHT];
  pad->nes_SELECT = keys[SDLK_SPACE];
  pad->nes_START = keys[SDLK_RETURN];
  pad->nes_B = keys[SDLK_z];
  pad->nes_A = keys[SDLK_x];
}


void
do_game (const char *filename, const char *geniefile, const char *nnnesterj_cheatfile) 
{
  NES *emu = NULL;
  sound_mgr *snd_mgr = NULL;
  int leftMargin = (screen->w - 256) >> 1;
  int topMargin = (screen->h - 224) >> 1;
  uint16 *fb = (uint16 *)screen->pixels;
  fb += topMargin * screen->pitch / 2;
  fb += leftMargin - 8;

#ifdef __UNIX__
  snd_mgr = new sdl_sound_mgr ();
#else
  snd_mgr = new null_sound_mgr ();
#endif
  if (!snd_mgr) goto finish;
  if (!(snd_mgr->initialize ())) goto finish;

  emu = new NES (snd_mgr);
  if (!emu) goto finish;
  g_emu = emu;
//   snprintf(emu->disksys_rom_filename, sizeof(emu->disksys_rom_filename),
//            "%s/disksys.rom", nes_homedir);
  snprintf(emu->disksys_rom_filename, sizeof(emu->disksys_rom_filename),
           "%s", rc_getstr("disksysrom"));
  if (!(emu->initialize (filename))) goto finish;

  sram_load (emu);
  disk_load (emu);

  exit_game_loop = false;
  interrupt_game_loop = false;
  frame_count = 0;

  if (geniefile)
    load_genie_from_file (geniefile, emu);

  if (nnnesterj_cheatfile)
    emu->load_nnnesterj_cheat (nnnesterj_cheatfile);

  printf("mapper = %d\n", emu->get_mapper_num());

  while (!exit_game_loop)
  {
    while (!interrupt_game_loop)
    {
#ifdef __SDL_DEBUG__
      if (draw_pattern)
	draw_pattern_table (emu);
#endif

      if (SDL_MUSTLOCK (screen)) SDL_LockSurface (screen);
      if (frame_count & 1)
        emu->emulate_frame (fb);
      else
        emu->emulate_frame_skip();

      {
	int y;
	uint16 *p;
	
#if 0
	for (y = 0; y < 224; ++y)
	{
	  p = (uint16 *)(screen->pixels) + 320 * y;
	  
	  *p++ = 0; *p++ = 0; *p++ = 0; *p++ = 0;
	  *p++ = 0; *p++ = 0; *p++ = 0; *p++ = 0;
	  
	  p += 256;
	  
	  *p++ = 0; *p++ = 0; *p++ = 0; *p++ = 0;
	  *p++ = 0; *p++ = 0; *p++ = 0; *p++ = 0;
	}
#endif
      }
      if (SDL_MUSTLOCK (screen)) SDL_UnlockSurface (screen);

      if (skip_sound)
	emu->emulate_frame_skip();

      {
        static int skinBlitCount = 0;
        if (skinBlitCount ++ < 2) {
          SDL_Rect r = { 0, 0, 320, 240 };
          SDL_FillRect(screen, &r, 0);
          SDL_Flip(screen);
        }
      }

//       SDL_Flip(screen);
      if (frame_count & 1) SDL_UpdateRect(screen, leftMargin, topMargin, 256, 224);

      ++frame_count;

//       poll_input(emu);
      doevents();
    }
  }

  fprintf(stderr, "1\n");
  sram_save (emu);
  fprintf(stderr, "2\n");
  disk_save (emu);
  fprintf(stderr, "3\n");

finish:

  fprintf(stderr, "4\n");
  if (emu) delete (emu);
  fprintf(stderr, "5\n");
  if (snd_mgr) delete (snd_mgr);
  fprintf(stderr, "6\n");
}


int
init_SDL_video ()
{
  const SDL_VideoInfo *vinf;
  int flags = 0;

  vinf = SDL_GetVideoInfo();

  if (vinf->hw_available)
    flags |= SDL_HWSURFACE;
//   flags |= SDL_DOUBLEBUF;

#ifdef __SDL_DEBUG__
  screen = SDL_SetVideoMode (320, 480, 16, flags);
#else
//   screen = SDL_SetVideoMode (320, 240, 16, flags);
  screen = SDL_SetVideoMode (320, 240, 16, flags);
#endif
  if (!screen) return 0;
  SDL_Flip(screen);

  if (rc_getint("grabint"))
    SDL_WM_GrabInput(SDL_GRAB_ON);

  SDL_WM_SetCaption ("NesterDC for SDL", "NesterDC for SDL");

  return 1;
}


int
shutdown_SDL_video ()
{
  SDL_Quit ();
}


#if 1
#include "testing.cpp"
#endif

#define DOTDIR ".znester"

static void
sys_initpath(void)
{
  char *buf, *home = getenv("HOME");
  mode_t mode = S_IRWXU;

  if (!home)
  {
    buf = ".";
    rc_setvar("rcpath", 1, &buf);
    rc_setvar("savedir", 1, &buf);
    rc_setvar("snapdir", 1, &buf);
    return;
  }

  buf = (char *)malloc(strlen(home) + strlen(DOTDIR) + 16);
  sprintf(buf, "%s/" DOTDIR, home);
  mkdir(buf, mode);

  sprintf(buf, "%s/" DOTDIR ":.", home);
  rc_setvar("rcpath", 1, &buf);

  sprintf(buf, "%s/" DOTDIR "/sav" , home);
  rc_setvar("savedir", 1, &buf);
  mkdir(buf, mode);

  sprintf(buf, "%s/" DOTDIR "/snap" , home);
  rc_setvar("snapdir", 1, &buf);
  mkdir(buf, mode);

  sprintf(buf, "%s/" DOTDIR "/disksys.rom", home);
  rc_setvar("disksysrom", 1, &buf);

  free(buf);
}

static char *
base(char *s)
{
  char *p;
  p = strrchr(s, '/');
  if (p) return p+1;
  return s;
}

extern "C" int
main(int orig_argc, char **orig_argv)
{
  int argc;
  char *argv[128];

  char module_dirname[512];
  char etc_dirname[512];
  int i;
  const char *filename = 0;

  if (mergeCmdlineFile(CMDLINE_FILE, orig_argc, orig_argv,
                       &argc, argv))
  {
    fprintf(stderr, "cmdline-file ('%s') is found.\n", CMDLINE_FILE);
    fprintf(stderr, "merge cmdline-file to shell arguments.\n");
  }

  init_exports();
  sys_initpath();

  for (i = 0; defaultconfig[i]; ++i)
    rc_command(defaultconfig[i]);

  i = parseArgs(argc, argv);
  if (i < 0)
    usage(base(argv[0]));
  filename = argv[i];

  keymap_sdl2dc_init ();

  if(SDL_Init(SDL_INIT_VIDEO)) //|SDL_INIT_AUDIO)==-1)
  {
    printf("Could not initialize SDL: %s.\n", SDL_GetError());
    return(-1);
  }

  if (!init_SDL_video ())
  {
    fprintf(stderr, "error in SDL video init\n");
    exit(1);
  }

#if 0
  test_all ();
#endif

  if (1)//(argc == 2)
    do_game (argv[1], NULL, NULL);
  else
  {
#if 0
    do_game (argv[1], argv[2], NULL);
#else
    do_game (filename, NULL, argv[2]);
#endif
  }

  shutdown_SDL_video ();
  return 0;
}


static int
mergeCmdlineFile(const char *filename, int orig_argc, char **orig_argv,
                 int *argc, char **argv)
{
  FILE *fp;
  char *cmdline;
  int ret = 0;

  *argc = 1;
  argv[0] = orig_argv[0];

  fp = fopen(filename, "rb");
  if (fp) {
    int filesize;
    char *p;

    fseek(fp, 0, SEEK_END);
    filesize = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    cmdline = (char *)malloc(filesize + 1); /* memory leak, but no problem. */

    if (fread(cmdline, 1, filesize, fp) != filesize)
    {
      fclose(fp);
      return 0;
    }
    fclose(fp);
    cmdline[filesize] = '\0';

    // Replace CR and LF characters to whitespace
    for (p = cmdline; *p != '\0'; ++p)
    {
      if (*p == '\r' || *p == '\n')
        *p = ' ';
    }

    *argc += splitline(argv + 1, 128 - orig_argc - 1, cmdline);
    ret = !0;
  }

  memcpy(argv + *argc, orig_argv + 1, sizeof(char *) * orig_argc);
  *argc += orig_argc - 1;

  return ret;
}

static int
parseArgs(int argc, char **argv)
{
  int ret = -1;
  char *cmd, *opt, *s, *arg;
  int i;

  for (i = 1; i < argc; i++)
  {
    if (!strcmp(argv[i], "--help"))
      help(base(argv[0]));
    else if (!strcmp(argv[i], "--version"))
      version(base(argv[0]));
    else if (!strcmp(argv[i], "--copying"))
      copying();
    else if (!strcmp(argv[i], "--showvars"))
    {
      show_exports();
      exit(0);
    }
    else if (!strcmp(argv[i], "--bind"))
    {
      if (i + 2 >= argc) {
        fprintf(stderr, "missing arguments to bind\n");
        exit(0);
      }
      cmd = (char *)malloc(strlen(argv[i+1]) + strlen(argv[i+2]) + 9);
      sprintf(cmd, "bind %s \"%s\"", argv[i+1], argv[i+2]);
      rc_command(cmd);
      free(cmd);
      i += 2;
    }
    else if (!strcmp(argv[i], "--source"))
    {
      if (i + 1 >= argc) {
        fprintf(stderr, "missing argument to source\n");
        exit(0);
      }
      cmd = (char *)malloc(strlen(argv[i+1]) + 8);
      sprintf(cmd, "source %s", argv[++i]);
      rc_command(cmd);
      free(cmd);
    }
    else if (!strncmp(argv[i], "--no-", 5))
    {
      opt = strdup(argv[i]+5);
      while ((s = strchr(opt, '-')))
        *s = '_';
      cmd = (char *)malloc(strlen(opt) + 7);
      sprintf(cmd, "set %s 0", opt);
      rc_command(cmd);
      free(cmd);
      free(opt);
    }
    else if (argv[i][0] == '-' && argv[i][1] == '-')
    {
      opt = strdup(argv[i]+2);
      if ((s = strchr(opt, '=')))
      {
        *s = 0;
        arg = s+1;
      }
      else
        arg = "1";
      while ((s = strchr(opt, '-')))
        *s = '_';
      while ((s = strchr(arg, ',')))
        *s = ' ';

      cmd = (char *)malloc(strlen(opt) + strlen(arg) + 6);
      sprintf(cmd, "set %s %s", opt, arg);

      rc_command(cmd);
      free(cmd);
      free(opt);
    }
    /* short options not yet implemented */
    else if (argv[i][0] == '-' && argv[i][1])
      ;
    else
      ret = i;
  }

  return ret;
}

void ev_poll(void);
void doevents(void)
{
  event_t ev;
  int st;

  ev_poll();

  while (ev_getevent(&ev))
  {
    if (ev.type != EV_PRESS && ev.type != EV_RELEASE)
      continue;
    st = (ev.type != EV_RELEASE);
    rc_dokey(ev.code, st);
  }
}

/* keymap - mappings of the form { scancode, localcode } - from sdl/keymap.c */
extern "C" int keymap[][2];

static int mapscancode(SDLKey sym)
{
  /* this could be faster:  */
  /*  build keymap as int keymap[256], then ``return keymap[sym]'' */

  int i;
  for (i = 0; keymap[i][0]; i++)
    if (keymap[i][0] == sym)
      return keymap[i][1];
  if (sym >= '0' && sym <= '9')
    return sym;
  if (sym >= 'a' && sym <= 'z')
    return sym;
  return 0;
}

void ev_poll(void)
{
  event_t ev;
  SDL_Event event;
  int axisval;

  static int prev_pen_axis[2] = { 0, 0 };

  while (SDL_PollEvent(&event))
  {
    switch(event.type)
    {
      case SDL_ACTIVEEVENT:
//         if (event.active.state == SDL_APPACTIVE)
//           fb.enabled = event.active.gain;
        break;
      case SDL_KEYDOWN:
        if ((event.key.keysym.sym == SDLK_RETURN) && (event.key.keysym.mod & KMOD_ALT))
          SDL_WM_ToggleFullScreen(screen);
        ev.type = EV_PRESS;
        ev.code = mapscancode(event.key.keysym.sym);
        ev_postevent(&ev);
        break;
      case SDL_KEYUP:
        ev.type = EV_RELEASE;
        ev.code = mapscancode(event.key.keysym.sym);
        ev_postevent(&ev);
        break;
      case SDL_MOUSEBUTTONDOWN:
        if ((screen->w >> 3) * 5 < event.button.x)
        {
          ev.type = EV_PRESS;
          prev_pen_axis[0] = ev.code = K_PENRIGHT;
          ev_postevent(&ev);
        }
        else if ((screen->w >> 3) * 3 > event.button.x)
        {
          ev.type = EV_PRESS;
          prev_pen_axis[0] = ev.code = K_PENLEFT;
          ev_postevent(&ev);
        }
        if ((screen->h >> 3) * 5 < event.button.y)
        {
          ev.type = EV_PRESS;
          prev_pen_axis[1] = ev.code = K_PENDOWN;
          ev_postevent(&ev);
        }
        else if ((screen->h >> 3) * 3 > event.button.y)
        {
          ev.type = EV_PRESS;
          prev_pen_axis[1] = ev.code = K_PENUP;
          ev_postevent(&ev);
        }

        if (!(prev_pen_axis[0] || prev_pen_axis[1]))
        {
          if ((screen->w >> 1) > event.button.x)
          {
            ev.type = EV_PRESS;
            prev_pen_axis[0] = ev.code = K_PENCENTERLEFT;
            ev_postevent(&ev);
          }
          else
          {
            ev.type = EV_PRESS;
            prev_pen_axis[0] = ev.code = K_PENCENTERRIGHT;
            ev_postevent(&ev);
          }
        }

        break;
      case SDL_MOUSEBUTTONUP:
        if (prev_pen_axis[0]) {
          ev.type = EV_RELEASE;
          ev.code = prev_pen_axis[0];
          prev_pen_axis[0] = 0;
          ev_postevent(&ev);
        }
        if (prev_pen_axis[1]) {
          ev.type = EV_RELEASE;
          ev.code = prev_pen_axis[1];
          prev_pen_axis[1] = 0;
          ev_postevent(&ev);
        }

        break;
      case SDL_QUIT:
        exit(1);
        break;
      default:
        break;
    }
  }
}
